from __future__ import annotations

import attrs
import numpy as np
from numpy.typing import NDArray

from ..base import LossBase
from .regression import L1Loss, L2Loss


@attrs.define()
class AsymmetricCompositeLoss(LossBase):
    """Asymmetric composite loss function.
    The loss function is `loss_pred_less` if `y_true < y_pred`,
    otherwise `loss_pred_greater`.
    """

    loss_pred_less: LossBase
    """The loss function if `y_true < y_pred`."""
    loss_pred_greater: LossBase
    """The loss function if `y_true >= y_pred`."""

    def loss(self, y_true: NDArray, y_pred: NDArray) -> float | NDArray:
        loss = np.where(
            y_true < y_pred,
            self.loss_pred_less.loss(y_true=y_true, y_pred=y_pred),
            self.loss_pred_greater.loss(y_true=y_true, y_pred=y_pred),
        )
        return loss

    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        grad, hess = np.where(
            y_true < y_pred,
            self.loss_pred_less.grad_hess(y_true=y_true, y_pred=y_pred),
            self.loss_pred_greater.grad_hess(y_true=y_true, y_pred=y_pred),
        )
        return grad, hess


class AsymmetricLoss(AsymmetricCompositeLoss):
    """Asymmetric loss function.
    The loss function is `loss * (1 - t)` if `y_true < y_pred`, otherwise `loss * t`.
    Generalized from quantile loss (pinball loss, check loss, etc.) and expectile loss.
    """

    def __init__(self, loss: LossBase, t: float = 0.5) -> None:
        super().__init__(loss * (1 - t), loss * t)


class QuantileLoss(AsymmetricLoss):
    """[Quantile](https://en.wikipedia.org/wiki/Quantile) loss function."""

    def __init__(self, t: float = 0.5) -> None:
        super().__init__(L1Loss(), t)


class ExpectileLoss(AsymmetricLoss):
    r"""[Expectile](https://sites.google.com/site/csphilipps/expectiles) loss function.
    - Expectile is a conditional mean if observations in [μ_τ, ∞)
    are τ/(1 - τ) times more likely than the original distribution.
    - Expectiles are not always quantiles of F.
    - Expectile loss is more smooth than quantile loss.

    .. math::
        \tau \int_{-\infty}^{\mu_\tau} (y - \mu_\tau) dF(y)
        = (1 - \tau) \int_{\mu_\tau}^{\infty} (y - \mu_\tau) dF(y)"""

    def __init__(self, t: float = 0.5) -> None:
        super().__init__(L2Loss(), t)



from .asymmetric import (
    AsymmetricCompositeLoss,
    AsymmetricLoss,
    ExpectileLoss,
    QuantileLoss,
)
from .regression import (
    FairLoss,
    GammaLoss,
    HuberLoss,
    L1Loss,
    L2Loss,
    LNLoss,
    LogCoshLoss,
    LogLoss,
    MAPELoss,
    MSLELoss,
    PoissonLoss,
    SMAPELoss,
    TweedieLoss,
)
from .sklearn import VarianceEstimator

__all__ = [
    "AsymmetricLoss",
    "AsymmetricCompositeLoss",
    "QuantileLoss",
    "ExpectileLoss",
    "L1Loss",
    "L2Loss",
    "LNLoss",
    "LogCoshLoss",
    "HuberLoss",
    "FairLoss",
    "PoissonLoss",
    "TweedieLoss",
    "GammaLoss",
    "LogLoss",
    "MAPELoss",
    "MSLELoss",
    "SMAPELoss",
    "VarianceEstimator",
]



from __future__ import annotations

from logging import getLogger

import attrs
from numpy.typing import NDArray

from .base import LossBase

LOG = getLogger(__name__)


@attrs.define()
class DebugLoss(LossBase):
    """Calls LOG.debug() every time loss() or grad_hess() is called."""

    loss_: LossBase

    def loss(self, y_true: NDArray, y_pred: NDArray) -> float | NDArray:
        loss = self.loss_.loss(y_true=y_true, y_pred=y_pred)
        LOG.debug(f"y_true: {y_true}, y_pred: {y_pred}, loss: {loss}")
        return loss

    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        grad, hess = self.loss_.grad_hess(y_true=y_true, y_pred=y_pred)
        LOG.debug(f"y_true: {y_true}, y_pred: {y_pred}, grad: {grad}, hess: {hess}")
        return grad, hess


@attrs.define()
class PrintLoss(LossBase):
    """Prints every time loss() or grad_hess() is called."""

    loss_: LossBase

    def loss(self, y_true: NDArray, y_pred: NDArray) -> float | NDArray:
        loss = self.loss_.loss(y_true=y_true, y_pred=y_pred)
        print(f"y_true: {y_true}, y_pred: {y_pred}, loss: {loss}")
        return loss

    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        grad, hess = self.loss_.grad_hess(y_true=y_true, y_pred=y_pred)
        print(f"y_true: {y_true}, y_pred: {y_pred}, grad: {grad}, hess: {hess}")
        return grad, hess



__version__ = "0.1.1"
from .base import LossBase
from .debug import DebugLoss, PrintLoss
from .resuming import ResumingLoss
from .sklearn import apply_custom_loss, patch_catboost

try:
    from .sklearn import patch_ngboost
except ImportError:
    pass
__all__ = [
    "LossBase",
    "DebugLoss",
    "PrintLoss",
    "ResumingLoss",
    "apply_custom_loss",
    "patch_catboost",
    "patch_ngboost",
]



from __future__ import annotations

from copy import copy
from typing import Any, Literal, Sequence

import numpy as np
from joblib import Parallel, delayed
from numpy.typing import NDArray
from sklearn.base import BaseEstimator
from typing_extensions import Self

from ..base import LossBase
from ..sklearn import apply_custom_loss
from .asymmetric import AsymmetricLoss


def _recursively_set_random_state(estimator: BaseEstimator, random_state: int) -> None:
    if hasattr(estimator, "random_state") and hasattr(estimator, "set_params"):
        estimator.set_params(random_state=random_state)
    for _, v in copy(estimator.get_params(deep=False)).items():
        if hasattr(v, "get_params"):
            _recursively_set_random_state(v, random_state)


class VarianceEstimator(BaseEstimator):
    """Estimator that estimates the distribution by simply using multiple estimators
    with different `t`.
    Compared to [NGBoost](https://stanfordmlgroup.github.io/projects/ngboost/) or
    [CatBoost's Uncertainty](https://catboost.ai/en/docs/references/uncertainty),
    this estimator is much slower and does not support "natural gradient",
    but does not require any assumption on the distribution.

    Note that NGBoost supports
    [any user-defineddistribution](https://stanfordmlgroup.github.io/ngboost/5-dev.html) # noqa
    but it has to be defined beforehand.

    NGBoost requires mean estimator and log standard deviation estimator
    to be trained simultaneously, which is very difficult to implement
    in sklearn / lightgbm / xgboost. (Need to start and stop fitting per iteration)
    Consider change `Base` parameter in NGBoost.
    (See https://github.com/stanfordmlgroup/ngboost/issues/250)
    """

    ts_: Sequence[float]
    m_type: Literal["mean", "median"]
    var_type: Literal["var", "std", "range", "mae", "mse"]

    def __init__(
        self,
        estimator: Any,
        loss: LossBase,
        *,
        ts: int | Sequence[float],
        n_jobs: int | None = 1,
        verbose: int = 0,
        random_state: int | None = None,
        m_type: Literal["mean", "median"] = "median",
        var_type: Literal["var", "std", "range", "mae", "mse"] = "var",
        target_transformer: BaseEstimator | Any | None = None,
    ) -> None:
        """Estimator that estimates the distribution by simply using multiple estimators
        with different `t`.
        Compared to [NGBoost](https://stanfordmlgroup.github.io/projects/ngboost/) or
        [CatBoost's Uncertainty](https://catboost.ai/en/docs/references/uncertainty),
        this estimator is much slower and does not support "natural gradient",
        but does not require any assumption on the distribution.

        Note that NGBoost supports
        [any user-defineddistribution](https://stanfordmlgroup.github.io/ngboost/5-dev.html) # noqa
        but it has to be defined beforehand.

        NGBoost requires mean estimator and log standard deviation estimator
        to be trained simultaneously, which is very difficult to implement
        in sklearn / lightgbm / xgboost. (Need to start and stop fitting per iteration)
        Consider change `Base` parameter in NGBoost.
        (See https://github.com/stanfordmlgroup/ngboost/issues/250)

        Parameters
        ----------
        estimator : Any
            The base estimator to use for fitting the data.
        loss : LossBase
            The loss function to use for fitting the data.
            Generally, `loss` should not be `AsymmetricLoss`.
        ts : int | Sequence[float]
            The list of `t` to use for fitting the data or the number of `t` to use.
            If `ts` is an integer, `np.linspace(1 / (ts * 2), 1 - 1 / (ts * 2), ts)` is used.
        n_jobs : int | None, optional
            The number of jobs to run in parallel for `fit`. `None` means 1.
        verbose : int, optional
            The verbosity level.
        random_state : int | None, optional
            The random state to use for fitting the data. If `None`, the random state is not set.
            If not `None`, new random state generated from `random_state` is set to each estimator.
        m_type : Literal[&quot;mean&quot;, &quot;median&quot;], optional
            M-statistics type to return from `predict` by default, by default "median"
        var_type : Literal[&quot;var&quot;, &quot;std&quot;, &quot;range&quot;, &quot;mae&quot;, &quot;mse&quot;], optional
            Variance type to return from `predict` by default, by default "var"
        target_transformer : BaseEstimator | Any | None, optional
            The transformer to use for transforming the target, by default None
            If `None`, no `TransformedTargetRegressor` is used.

        Raises
        ------
        TypeError
            Raises if `estimator` does not have `fit` method or `predict` method.
        """
        if not hasattr(estimator, "fit"):
            raise TypeError(f"{estimator} does not have fit method")
        if not hasattr(estimator, "predict"):
            raise TypeError(f"{estimator} does not have predict method")
        self.estimator = estimator
        self.loss = loss
        self.ts = ts
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.random_state = random_state
        self.m_type = m_type
        self.var_type = var_type
        self.target_transformer = target_transformer
        self.random = np.random.RandomState(random_state)

    def fit(self, X: Any, y: Any, **fit_params: Any) -> Self:
        """Fit each estimator with different `t`.

        Parameters
        ----------
        X : Any
            The training input samples.
        y : Any
            The target values.

        Returns
        -------
        Self
            Fitted estimator.

        Raises
        ------
        RuntimeError
            Raises if joblib fails to return the results.
        """
        ts = self.ts
        if isinstance(ts, int):
            ts = np.linspace(1 / (ts * 2), 1 - 1 / (ts * 2), ts)
        self.ts_ = ts  # type: ignore
        estimators_ = [
            apply_custom_loss(
                self.estimator,
                AsymmetricLoss(self.loss, t=t),
                target_transformer=self.target_transformer,
            )
            for t in self.ts_
        ]
        if self.random is not None:
            for estimator in estimators_:
                _recursively_set_random_state(
                    estimator, self.random.randint(0, np.iinfo(np.int32).max)
                )
        parallel_result = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)(
            [delayed(estimator.fit)(X, y, **fit_params) for estimator in estimators_]
        )
        if parallel_result is None:
            raise RuntimeError("joblib.Parallel returned None")
        self.estimators_ = parallel_result
        return self

    def predict_raw(self, X: Any, **predict_params: Any) -> NDArray[Any]:
        """Returns raw predictions of each estimator.

        Parameters
        ----------
        X : Any
            X
        **predict_params : Any
            The parameters to be passed to `predict` method of each estimator.

        Returns
        -------
        NDArray[Any]
            Raw predictions of each estimator with shape (n_estimators, n_samples)
        """
        return np.array(
            [estimator.predict(X, **predict_params) for estimator in self.estimators_]
        )

    def predict(
        self,
        X: Any,
        type_: Literal["mean", "median", "var", "std", "range", "mae", "mse"]
        | None = None,
        **predict_params: Any,
    ) -> NDArray[Any]:
        """Returns predictions of the ensemble.

        Parameters
        ----------
        X : Any
            X
        type_ : Literal['mean', 'median', 'var', 'std', 'range', 'mae', 'mse'], optional
            Type of the prediction, by default None
            If None, self.m_type is used.
        **predict_params : Any
            The parameters to be passed to `predict` method of each estimator.

        Returns
        -------
        NDArray[Any]
            Predictions of the ensemble with shape (n_samples,)

        Raises
        ------
        ValueError
            When type_ is not supported.
        """
        type_ = type_ or self.m_type
        if type_ == "mean":
            return self.predict_raw(X, **predict_params).mean(axis=0)
        elif type_ == "median":
            return np.median(self.predict_raw(X, **predict_params), axis=0)
        elif type_ == "var":
            return self.predict_raw(X, **predict_params).var(axis=0)
        elif type_ == "std":
            return self.predict_raw(X, **predict_params).std(axis=0)
        elif type_ == "range":
            return self.predict_raw(X, **predict_params).max(axis=0) - self.predict_raw(
                X, **predict_params
            ).min(axis=0)
        elif type_ == "mae":
            return np.abs(
                self.predict_raw(X, **predict_params)
                - self.predict_raw(X, **predict_params).mean(axis=0)
            ).mean(axis=0)
        elif type_ == "mse":
            return (
                (
                    self.predict_raw(X, **predict_params)
                    - self.predict_raw(X, **predict_params).mean(axis=0)
                )
                ** 2
            ).mean(axis=0)
        else:
            raise ValueError(f"Unknown type_: {type_}")

    def predict_var(
        self,
        X: Any,
        type_: Literal["var", "std", "range", "mae", "mse"] | None = None,
        **predict_params: Any,
    ) -> NDArray[Any]:
        """Returns variance of the ensemble.

        Parameters
        ----------
        X : Any
            X
        type_ : Literal['var', 'std', 'range', 'mae', 'mse'], optional
            Type of the variance, by default None
            If None, self.var_type is used.
        **predict_params : Any
            The parameters to be passed to `predict` method of each estimator.

        Returns
        -------
        NDArray[Any]
            Variance of the ensemble with shape (n_samples,)
        """
        type_ = type_ or self.var_type
        return self.predict(X, type_=type_, **predict_params)



from __future__ import annotations

import functools
import importlib.util
from typing import Any, Literal, TypeVar, overload
from unittest.mock import patch

import catboost as cb
import lightgbm as lgb
import numpy as np
import xgboost as xgb
from sklearn.base import BaseEstimator, clone
from sklearn.compose import TransformedTargetRegressor
from sklearn.preprocessing import StandardScaler

from .base import LossBase

TEstimator = TypeVar("TEstimator", cb.CatBoost, lgb.LGBMModel, xgb.XGBModel)


@overload
def apply_custom_loss(
    estimator: TEstimator,
    loss: LossBase,
    *,
    copy: bool = ...,
    target_transformer: None = ...,
    recursive: bool = ...,
) -> TEstimator:
    ...


@overload
def apply_custom_loss(
    estimator: TEstimator,
    loss: LossBase,
    *,
    copy: bool = ...,
    target_transformer: BaseEstimator = ...,
    recursive: bool = ...,
) -> TransformedTargetRegressor:
    ...


def apply_custom_loss(
    estimator: TEstimator,
    loss: LossBase,
    *,
    copy: bool = True,
    target_transformer: BaseEstimator | Any | None = StandardScaler(),
    recursive: bool = True,
) -> TEstimator | TransformedTargetRegressor:
    """Apply custom loss to the estimator.

    Parameters
    ----------
    estimator : TEstimator
        CatBoost, LGBMModel, or XGBModel
    loss : LossBase
        The custom loss to apply
    copy : bool, optional
        Whether to copy the estimator using `sklearn.base.clone`, by default True
    target_transformer : BaseEstimator | Any | None, optional
        The target transformer to use, by default StandardScaler()
        (This option exists because some loss functions require the target
        to be normalized (i.e. `LogCoshLoss`))
    recursive : bool, optional
        Whether to recursively search for estimators inside the estimator
        and apply the custom loss to all of them, by default True

    Returns
    -------
    TEstimator | TransformedTargetRegressor
        The estimator with the custom loss applied
    """
    if copy:
        estimator = clone(estimator)
    if isinstance(estimator, cb.CatBoost):
        estimator.set_params(loss_function=loss, eval_metric=loss)
    if isinstance(estimator, lgb.LGBMModel):
        estimator.set_params(objective=loss)
        estimator_fit = estimator.fit

        @functools.wraps(estimator_fit)
        def fit(self: lgb.LGBMModel, X: Any, y: Any, **fit_params: Any) -> Any:
            fit_params["eval_metric"] = loss.eval_metric_lgb
            return estimator_fit(self, X, y, **fit_params)

        patch.object(estimator, "fit", fit)
    if isinstance(estimator, xgb.XGBModel):
        estimator.set_params(objective=loss, eval_metric=loss.eval_metric_xgb_sklearn)

    if recursive:
        for key, value in estimator.get_params(deep=True).items():
            if hasattr(value, "fit"):
                estimator.set_params(
                    **{
                        key: apply_custom_loss(
                            value, loss, copy=False, target_transformer=None
                        )
                    }
                )

    if target_transformer is None:
        return estimator
    return TransformedTargetRegressor(estimator, transformer=clone(target_transformer))


if importlib.util.find_spec("ngboost") is not None:
    from ngboost import NGBoost
    from ngboost.distns import Normal
    from numpy.typing import NDArray

    def patch_ngboost(estimator: NGBoost) -> NGBoost:
        """Patch NGBoost to return only the mean prediction in `predict`
        and the variance in `predict_var` to be consistent with other models.

        Parameters
        ----------
        estimator : NGBoost
            The NGBoost estimator to patch.

        Returns
        -------
        NGBoost
            The patched NGBoost estimator.
        """

        self = estimator

        def predict_var(X: Any, **predict_params: Any) -> NDArray[Any]:
            dist = self.pred_dist(X, **predict_params)
            if not isinstance(dist, Normal):
                raise NotImplementedError
            return dist.var

        setattr(estimator, "predict_var", predict_var)

        def predict_std(X: Any, **predict_params: Any) -> NDArray[Any]:
            dist = self.pred_dist(X, **predict_params)
            if not isinstance(dist, Normal):
                raise NotImplementedError
            return dist.scale

        setattr(estimator, "predict_std", predict_std)
        return estimator


def patch_catboost(estimator: cb.CatBoost) -> cb.CatBoost:
    """Patch CatBoost to return only the mean prediction in `predict`
    and the variance in `predict_var` to be consistent with other models.

    Parameters
    ----------
    estimator : cb.CatBoost
        The CatBoost estimator to patch.

    Returns
    -------
    cb.CatBoost
        The patched CatBoost estimator.
    """
    original_predict = estimator.predict

    def predict(
        data: Any,
        prediction_type: Literal[
            "Probability", "Class", "RawFormulaVal", "Exponent", "LogProbability"
        ] = "RawFormulaVal",
        ntree_start: int = 0,
        ntree_end: int = 0,
        thread_count: int = -1,
        verbose: bool | None = None,
        task_type: str = "CPU",
    ) -> NDArray[Any]:
        prediction = original_predict(
            data,
            prediction_type,
            ntree_start,
            ntree_end,
            thread_count,
            verbose,
            task_type,
        )
        if prediction.ndim == 2:
            return prediction[:, 0]
        return prediction

    setattr(estimator, "predict", predict)

    self = estimator

    def predict_var(
        X: Any,
        prediction_type: Literal["knowledge", "data", "total"] = "total",
        **predict_params: Any,
    ) -> NDArray[Any]:
        uncertainty = self.virtual_ensembles_predict(
            X, prediction_type="TotalUncertainty", **predict_params
        )

        loss_function = self.get_params()["loss_function"]
        knowledge_uncertainty = np.zeros_like(uncertainty[..., 0])
        data_uncertainty = np.zeros_like(uncertainty[..., 0])
        if loss_function == "RMSEWithUncertainty":
            knowledge_uncertainty = uncertainty[..., 1]
            data_uncertainty = uncertainty[..., 2]
        elif isinstance(estimator, cb.CatBoostClassifier):
            knowledge_uncertainty = uncertainty[..., 1]
        else:
            data_uncertainty = uncertainty[..., 0]
            knowledge_uncertainty = uncertainty[..., 1] - data_uncertainty

        if prediction_type == "knowledge":
            return knowledge_uncertainty
        elif prediction_type == "data":
            return data_uncertainty
        elif prediction_type == "total":
            return knowledge_uncertainty + data_uncertainty
        else:
            raise ValueError(
                "prediction_type must be one of ['knowledge', 'data', 'total'], "
                f"but got {prediction_type}"
            )

    setattr(estimator, "predict_var", predict_var)
    return estimator



from __future__ import annotations

import warnings
from abc import ABCMeta
from logging import getLogger
from numbers import Real
from typing import Any, Callable, Sequence, final

import attrs
import humps
import lightgbm as lgb
import numpy as np
import xgboost as xgb
from numpy.typing import NDArray
from typing_extensions import Self

LOG = getLogger(__name__)


def _dataset_to_ndarray(
    y: NDArray | lgb.Dataset | xgb.DMatrix,
) -> tuple[NDArray, NDArray]:
    if isinstance(y, lgb.Dataset):
        y_ = y.get_label()
        if y_ is None:
            raise ValueError("y is None")
        weight = y.get_weight()
        if weight is None:
            weight = np.ones_like(y_)
        return y_, weight
    if isinstance(y, xgb.DMatrix):
        y_ = y.get_label()
        return y_, np.ones_like(y_)
    return y, np.ones_like(y)


def _get_name_from_callable(obj: Callable[..., Any]) -> str:
    if hasattr(obj, "__name__"):
        return getattr(obj, "__name__")
    if hasattr(obj, "__class__") and hasattr(getattr(obj, "__class__"), "__name__"):
        return getattr(getattr(obj, "__class__"), "__name__")
    raise ValueError(f"Could not get name from callable {obj}")


class LossBase(metaclass=ABCMeta):
    """Base class for loss functions.
    Inherit this class to implement custom loss function.

    See Also
    --------
    Catboost:
    https://catboost.ai/en/docs/concepts/python-usages-examples#user-defined-loss-function

    LightGBM:
    https://lightgbm.readthedocs.io/en/latest/Advanced-Topics.html#custom-objective-function

    XGBoost:
    https://xgboost.readthedocs.io/en/latest/tutorials/custom_metric_obj.html

    Example
    -------
    >>> from boost_loss.base import LossBase
    >>> import numpy as np
    >>> from numpy.typing import NDArray
    >>>
    >>> class L2Loss(LossBase):
    >>>     def loss(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
    >>>         return (y_true - y_pred) ** 2
    >>>     def grad(self, y_true: NDArray, y_pred: NDArray) -> NDArray: # dL/dy_pred
    >>>         return -2 * (y_true - y_pred) # or (y_pred - y_true)
    >>>     def hess(self, y_true: NDArray, y_pred: NDArray) -> NDArray: # d2L/dy_pred2
    >>>         return 2 * np.ones_like(y_true) # or np.ones_like(y_true)
    >>>
    >>> from boost.sklearn import apply_custom_loss
    >>> import lightgbm as lgb
    >>> apply_custom_loss(lgb.LGBMRegressor(), L2Loss()).fit(X, y)
    """

    is_higher_better: bool = False
    """Whether the result of loss function is better when it is higher."""

    @classmethod
    @final
    def from_callable(
        cls,
        loss: Callable[[NDArray, NDArray], NDArray | float],
        grad: Callable[[NDArray, NDArray], NDArray],
        hess: Callable[[NDArray, NDArray], NDArray],
        name: str | None = None,
        is_higher_better: bool = False,
    ) -> type[Self]:
        """Create this class from loss, grad, and hess callables.

        Parameters
        ----------
        loss : Callable[[NDArray, NDArray], NDArray | float]
            The loss function. If 1-D array is returned,
            the mean of array is calculated.
            Return 1-D array if possible in order to utilize weights in the dataset
            if available.
            (y_true, y_pred) -> loss
        grad : Callable[[NDArray, NDArray], NDArray]
            The 1st order derivative (gradient) of loss w.r.t. y_pred.
            (y_true, y_pred) -> grad
        hess : Callable[[NDArray, NDArray], NDArray]
            The 2nd order derivative (Hessian) of loss w.r.t. y_pred.
            (y_true, y_pred) -> hess
        name : str | None, optional
            The name of loss function.
            If None, it tries to infer from loss function, by default None
        is_higher_better : bool, optional
            Whether the result of loss function is better when it is higher,
            by default False

        Returns
        -------
        type[Self]
            The subclass of this class.

        Raises
        ------
        ValueError
            If name is None and it can't infer from loss function.
        """
        if name is None:
            try:
                name = _get_name_from_callable(loss)
            except ValueError as e:
                raise ValueError(
                    "Could not infer name from loss function. Please specify name."
                ) from e
        return type(
            name,
            (cls,),
            dict(
                loss=staticmethod(loss),
                grad=staticmethod(grad),
                hess=staticmethod(hess),
                is_higher_better=is_higher_better,
            ),
        )

    @property
    def _grad_hess_sign(self) -> int:
        return -1 if self.is_higher_better else 1

    def __init_subclass__(cls, **kwargs: Any) -> None:
        grad_inherited = cls.grad is not LossBase.grad
        hess_inherited = cls.hess is not LossBase.hess
        grad_hess_inherited = cls.grad_hess is not LossBase.grad_hess
        if grad_inherited and hess_inherited:
            pass
        elif grad_hess_inherited:
            pass
        else:
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} "
                "with grad_hess or both grad and hess not implemented"
            )
        super().__init_subclass__(**kwargs)

    @property
    def name(self) -> str:
        """Name of loss function.

        Returns
        -------
        str
            Snake case of class name. e.g. `LogCoshLoss` -> `log_cosh_loss`.
        """
        return humps.decamelize(self.__class__.__name__.replace("Loss", ""))

    def grad(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
        """The 1st order derivative (gradient) of loss w.r.t. y_pred.

        Parameters
        ----------
        y_true : NDArray
            The true target values.
        y_pred : NDArray
            The predicted target values.

        Returns
        -------
        NDArray
            The gradient of loss function. 1-D array with shape (n_samples,).

        Raises
        ------
        NotImplementedError
            If not implemented.
        """
        raise NotImplementedError()

    def hess(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
        """The 2nd order derivative (hessian) of loss w.r.t. y_pred.

        Parameters
        ----------
        y_true : NDArray
            The true target values.
        y_pred : NDArray
            The predicted target values.

        Returns
        -------
        NDArray
            The hessian of loss function. 1-D array with shape (n_samples,).

        Raises
        ------
        NotImplementedError
            If not implemented.
        """
        raise NotImplementedError()

    def loss(self, y_true: NDArray, y_pred: NDArray) -> NDArray | float:
        """Loss function. If 1-D array is returned, the mean of array is calculated.
        Return 1-D array if possible in order to utilize weights in the dataset
        if available.

        Parameters
        ----------
        y_true : NDArray
            The true target values.
        y_pred : NDArray
            The predicted target values.

        Returns
        -------
        NDArray | float
            The loss function. 1-D array with shape (n_samples,) or float.

        Raises
        ------
        NotImplementedError
            If not implemented.
        """
        raise NotImplementedError()

    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        """Gradient and hessian of loss function. Override this method if you want to
        calculate both gradient and hessian at the same time.

        Parameters
        ----------
        y_true : NDArray
            The true target values.
        y_pred : NDArray
            The predicted target values.

        Returns
        -------
        tuple[NDArray, NDArray]
            The gradient and hessian of loss function.
            1-D array with shape (n_samples,).
        """
        return self.grad(y_true=y_true, y_pred=y_pred), self.hess(
            y_true=y_true, y_pred=y_pred
        )

    def _grad_hess_weighted(
        self, y_true: NDArray, y_pred: NDArray, weight: NDArray
    ) -> tuple[NDArray, NDArray]:
        grad, hess = self.grad_hess(y_true=y_true, y_pred=y_pred)
        if np.any(hess < 0):
            negative_rate = np.mean(hess < 0)
            warnings.warn(
                f"Found negative hessian in {negative_rate:.2%} samples."
                "This may cause convergence issue and cause CatBoostError in CatBoost."
                "If LightGBM or XGBoost is used, the estimator will return "
                "nonsense values (like all 0s if 100%).",
                RuntimeWarning,
            )
        grad, hess = grad * weight, hess * weight
        grad, hess = grad * self._grad_hess_sign, hess * self._grad_hess_sign
        return grad, hess

    @final
    def __call__(
        self,
        y_true: NDArray | lgb.Dataset | xgb.DMatrix,
        y_pred: NDArray | lgb.Dataset | xgb.DMatrix,
    ) -> tuple[NDArray, NDArray]:
        """Sklearn-compatible interface (Sklearn, LightGBM, XGBoost)"""
        if isinstance(y_pred, lgb.Dataset) or isinstance(y_pred, xgb.DMatrix):
            # NOTE: swap (it is so fucking that the order is inconsistent)
            y_true, y_pred = y_pred, y_true
        y_true, weight = _dataset_to_ndarray(y=y_true)
        y_pred, _ = _dataset_to_ndarray(y=y_pred)
        return self._grad_hess_weighted(y_true=y_true, y_pred=y_pred, weight=weight)

    @final
    def calc_ders_range(
        self,
        preds: Sequence[float],
        targets: Sequence[float],
        weights: Sequence[float] | None = None,
    ) -> list[tuple[float, float]]:
        """Catboost-compatible interface"""
        y_pred = np.array(preds)
        y_true = np.array(targets)
        weight = np.array(weights) if weights is not None else np.ones_like(y_pred)
        grad, hess = self._grad_hess_weighted(
            y_true=y_true, y_pred=y_pred, weight=weight
        )
        # NOTE: in catboost, the definition of loss is the inverse
        grad, hess = -grad, -hess
        return list(zip(grad, hess))

    @final
    def is_max_optimal(self) -> bool:
        """Catboost-compatible interface"""
        return self.is_higher_better

    @final
    def evaluate(
        self,
        approxes: Sequence[float],
        target: Sequence[float],
        weight: Sequence[float] | None = None,
    ) -> tuple[float, float]:
        """Catboost-compatible interface"""
        approxes_ = np.array(approxes[0])
        targets_ = np.array(target)
        weights_ = np.array(weight) if weight is not None else np.ones_like(approxes_)
        loss = self.loss(y_true=targets_, y_pred=approxes_)
        if isinstance(loss, float) and not np.allclose(weights_, 1.0):
            warnings.warn("loss() should return ndarray when weight is not all 1.0")
            return loss, np.nan
        return float(np.sum(loss * weights_)), float(np.sum(weights_))

    @final
    def get_final_error(self, error: float, weight: float | None = None) -> float:
        """Catboost-compatible interface"""
        return error / (weight + 1e-38) if weight is not None else error

    @final
    def eval_metric_lgb(
        self,
        y_true: NDArray | lgb.Dataset | xgb.DMatrix,
        y_pred: NDArray | lgb.Dataset | xgb.DMatrix,
    ) -> tuple[str, float, bool]:
        """LightGBM-compatible interface"""
        if isinstance(y_pred, lgb.Dataset) or isinstance(y_pred, xgb.DMatrix):
            # NOTE: swap (it is so fucking that the order is inconsistent)
            y_true, y_pred = y_pred, y_true
        y_true, weight = _dataset_to_ndarray(y=y_true)
        y_pred, _ = _dataset_to_ndarray(y=y_pred)
        loss = self.loss(y_true=y_true, y_pred=y_pred)
        if isinstance(loss, float) and not np.allclose(weight, 1.0):
            warnings.warn("loss() should return ndarray when weight is not all 1.0")
            return self.name, loss, self.is_higher_better
        return (
            self.name,
            float(np.sum(loss * weight) / (np.sum(weight) + 1e-38)),
            self.is_higher_better,
        )

    @final
    def eval_metric_xgb_native(
        self,
        y_true: NDArray | lgb.Dataset | xgb.DMatrix,
        y_pred: NDArray | lgb.Dataset | xgb.DMatrix,
    ) -> tuple[str, float]:
        """XGBoost-native-api-compatible interface"""
        result = self.eval_metric_lgb(y_true=y_true, y_pred=y_pred)
        return result[0], result[1]

    @final
    def eval_metric_xgb_sklearn(
        self,
        y_true: NDArray | lgb.Dataset | xgb.DMatrix,
        y_pred: NDArray | lgb.Dataset | xgb.DMatrix,
    ) -> float:
        """XGBoost-sklearn-api-compatible interface"""
        result = self.eval_metric_lgb(y_true=y_true, y_pred=y_pred)
        return result[1]

    def __add__(self, other: LossBase) -> LossBase:
        if not isinstance(other, LossBase):
            return NotImplemented  # type: ignore
        return _LossSum(self, other)

    def __sub__(self, other: LossBase) -> LossBase:
        return self.__add__(-other)

    def __mul__(self, other: float | int | Real) -> LossBase:
        if not isinstance(other, Real):
            return NotImplemented
        return _LossMul(self, other)

    def __div__(self, other: float | int | Real) -> LossBase:
        return self.__mul__(1.0 / other)

    def __radd__(self, other: LossBase) -> LossBase:
        return self.__add__(other)

    def __rsub__(self, other: LossBase) -> LossBase:
        return self.__sub__(other)

    def __rmul__(self, other: float | int | Real) -> LossBase:
        return self.__mul__(other)

    def __rdiv__(self, other: float | int | Real) -> LossBase:
        return self.__div__(other)

    def __neg__(self) -> LossBase:
        return self.__mul__(-1.0)

    def __pos__(self) -> Self:
        return self


@attrs.define()
class _LossSum(LossBase):
    loss1: LossBase
    loss2: LossBase

    @property
    def name(self) -> str:
        return self.loss1.name + "+" + self.loss2.name

    def loss(self, y_true: NDArray, y_pred: NDArray) -> NDArray | float:
        return self.loss1.loss(y_true=y_true, y_pred=y_pred) + self.loss2.loss(
            y_true=y_true, y_pred=y_pred
        )

    def grad(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
        return self.loss1.grad(y_true=y_true, y_pred=y_pred) + self.loss2.grad(
            y_true=y_true, y_pred=y_pred
        )

    def hess(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
        return self.loss1.hess(y_true=y_true, y_pred=y_pred) + self.loss2.hess(
            y_true=y_true, y_pred=y_pred
        )

    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        grad1, hess1 = self.loss1.grad_hess(y_true=y_true, y_pred=y_pred)
        grad2, hess2 = self.loss2.grad_hess(y_true=y_true, y_pred=y_pred)
        return grad1 + grad2, hess1 + hess2


@attrs.define()
class _LossMul(LossBase):
    loss_: LossBase
    factor: float | int | Real

    @property
    def name(self) -> str:
        return f"{self.factor}*{self.loss_.name}"

    def loss(self, y_true: NDArray, y_pred: NDArray) -> NDArray | float:
        return self.factor * self.loss_.loss(y_true=y_true, y_pred=y_pred)

    def grad(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
        return self.factor * self.loss_.grad(y_true=y_true, y_pred=y_pred)

    def hess(self, y_true: NDArray, y_pred: NDArray) -> NDArray:
        return self.factor * self.loss_.hess(y_true=y_true, y_pred=y_pred)

    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        grad, hess = self.loss_.grad_hess(y_true=y_true, y_pred=y_pred)
        return self.factor * grad, self.factor * hess



from __future__ import annotations

from abc import ABCMeta
from typing import Any, Callable, final

import attrs
import numpy as np
import torch
from numpy.typing import NDArray
from typing_extensions import Self

from .base import LossBase, _get_name_from_callable


class TorchLossBase(LossBase, metaclass=ABCMeta):
    """Calculate gradient and hessian using `torch.autograd.grad`.
    One of `loss_torch` and `grad_torch` must be implemented.

    Inspired by
    https://github.com/TomerRonen34/treeboost_autograd/blob/main/treeboost_autograd/pytorch_objective.py # noqa: E501
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        loss_inherited = cls.loss_torch is not TorchLossBase.loss_torch
        grad_inherited = cls.grad_torch is not TorchLossBase.grad_torch
        if loss_inherited or grad_inherited:
            pass
        else:
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} "
                "with loss_torch or grad_torch not implemented"
            )
        return super().__init_subclass__(**kwargs)

    def loss_torch(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
        """The loss function.  If 1-D array is returned,
        the mean of array is calculated.
        Return 1-D array if possible in order to utilize weights in the dataset
        if available.

        Parameters
        ----------
        y_true : torch.Tensor
            The true target values.
        y_pred : torch.Tensor
            The predicted target values.

        Returns
        -------
        torch.Tensor
            0-dim or 1-dim tensor of shape (n_samples,). Return 1-dim tensor if possible
            to utilize weights in the dataset if available.

        Raises
        ------
        NotImplementedError
            If not implemented.
        """
        raise NotImplementedError()

    def grad_torch(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
        """The 1st order derivative of loss w.r.t. y_pred.

        Parameters
        ----------
        y_true : torch.Tensor
            The true target values.
        y_pred : torch.Tensor
            The predicted target values.

        Returns
        -------
        torch.Tensor
            The gradient of loss w.r.t. y_pred. 1-dim tensor of shape (n_samples,).

        Raises
        ------
        NotImplementedError
            If not implemented.
        """
        raise NotImplementedError()

    @final
    def loss(self, y_true: NDArray, y_pred: NDArray) -> NDArray | float:
        loss = self.loss_torch(
            torch.from_numpy(y_true), torch.from_numpy(y_pred)
        ).detach()
        if loss.ndim == 0:
            return loss.item()
        else:
            return loss.numpy()

    @final
    def grad_hess(self, y_true: NDArray, y_pred: NDArray) -> tuple[NDArray, NDArray]:
        y_true_ = torch.from_numpy(y_true)
        y_pred_ = torch.from_numpy(y_pred).requires_grad_(True)

        try:
            grad = self.grad_torch(y_true_, y_pred_)
        except NotImplementedError:
            loss = torch.mean(self.loss_torch(y_true_, y_pred_)) * y_true_.size(0)
            grad = torch.autograd.grad(loss, y_pred_, create_graph=True)[0]
            hess = np.array(
                [
                    torch.autograd.grad(grad_, y_pred_, retain_graph=True)[0][i].item()
                    for i, grad_ in enumerate(grad)
                ]
            )
        else:
            hess = np.array(
                [
                    torch.autograd.grad(grad_, y_pred_, create_graph=True)[0][i].item()
                    for i, grad_ in enumerate(grad)
                ]
            )
        return grad.detach().numpy(), hess

    @classmethod
    @final
    def from_callable_torch(
        cls,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        name: str | None = None,
        is_higher_better: bool = False,
    ) -> type[Self]:
        """Create a loss class from a callable.

        Parameters
        ----------
        loss : Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function.  If 1-D array is returned,
            the mean of array is calculated.
            Return 1-D array if possible in order to utilize weights in the dataset
            if available.
            (y_true, y_pred) -> loss
        name : str | None, optional
            The name of loss function.
            If None, it tries to infer from loss function, by default None
        is_higher_better : bool, optional
            Whether the result of loss function is better when it is higher,
            by default False

        Returns
        -------
        type[Self]
            The subclass of this class.

        Raises
        ------
        ValueError
            If name is None and it can't infer from loss function.
        """
        if name is None:
            try:
                name = _get_name_from_callable(loss)
            except ValueError as e:
                raise ValueError(
                    "Could not infer name from loss function. Please specify name."
                ) from e
        return type(
            name,
            (cls,),
            dict(loss_torch=staticmethod(loss), is_higher_better=is_higher_better),
        )


@attrs.define(kw_only=True)
class _LNLossTorch_(TorchLossBase):
    """L^n loss for PyTorch."""

    n: float
    divide_n_loss: bool = False
    divide_n_grad: bool = False

    def loss_torch(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
        if self.divide_n_grad != self.divide_n_loss:
            raise ValueError(
                "divide_n_grad and divide_n_loss must be the same, "
                f"but got {self.divide_n_grad} and {self.divide_n_loss}"
            )
        return torch.abs(y_true - y_pred) ** self.n / (
            self.n if self.divide_n_loss else 1
        )


@attrs.define(kw_only=True)
class _LNLossTorch(TorchLossBase):
    """L^n loss for PyTorch."""

    n: float
    divide_n_loss: bool = False
    divide_n_grad: bool = True

    def loss_torch(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
        return torch.abs(y_pred - y_true) ** self.n / (
            self.n if self.divide_n_loss else 1
        )

    def grad_torch(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
        return (
            torch.sign(y_pred - y_true)
            * torch.abs(y_pred - y_true) ** (self.n - 1)
            * self.n
            / (self.n if self.divide_n_grad else 1)
        )
